# $Name:  $ 
# $Author: dgp $ 
# $Date: 1997/03/22 00:23:47 $ 
#
# A library of Tcl procs which manipulate strings containing URLs, parsing
# them into their subparts, and making useful conversions.
#
# Need to add a UrlUnescape to undo UrlEscape and use them properly

# Add Url package to database; define standard global variables
OMFCVS_PackageProvide {^Name: Url-0_0a0^} url

# proc ParseUrl absoluteURL 
#       Parses the string absoluteURL as an absolulte URL according to
#       RFC 1738
proc OMFURL_ParseUrl {absoluteURL} {
  # URLs contain no white space.  If there is some, it was put in
  # by mistake, so get rid of it.
  regsub -all "\[ \t\r\n\]" $absoluteURL {} absoluteURL
  # Note: Although RFC1738 claims "~" is not a legal URL character,
  # and should be escaped, in practice this is not enforced, so we
  # don't either.
  if [regexp {^([a-zA-Z0-9+.-]+):([a-zA-Z0-9$_+.!*'(),;/?:@&=%~-]*)$} \
      $absoluteURL match URLparts(scheme) URLparts(schemepart) ] {
    set URLparts(scheme) [string tolower $URLparts(scheme)]
    return [array get URLparts]
  } else {
    return -code error -errorinfo "Missing protocol/scheme" !
  }
}

# proc ParseIpSchemepart URLpartlist
#       Parses the schemepart of a URL whose scheme makes use of the
#       IP schemepart syntax (see RFC 1738 and RFC 1808)
#
#       For generic-RLs,        IPSchemepart    ==      net_path
#                               (RFC 1738)              (RFC 1808)
#
#       Assumes that the schemepart has already been verified, say, by
#       parseURL
proc OMFURL_ParseIpSchemepart {URLpartlist} {
  array set URLparts $URLpartlist
  if [regexp {^//([^/]+)(/(.*))?$} $URLparts(schemepart) match \
      URLparts(net_loc) foo URLparts(rel_path) ] {
    while {[regsub ^/ $URLparts(rel_path) {} URLparts(rel_path)]} {}
    set URLparts(abs_path) /$URLparts(rel_path)
    return [array get URLparts]
  } else {
    return -code error -errorinfo "Invalid protocol syntax for \
protocol $URLparts(scheme)" /
  }
}

# proc MakeAbsoluteUrl relativeURL baseURL
#       Take a (possibly) relative URL, and convert it into absolute 
#       form using the (absolute or empty) baseURL as a reference.
#       Assumes baseURL is legal.  Checks relativeURL.
#       See RFC 1808.
proc OMFURL_MakeAbsoluteUrl {relativeURL baseURL} {
  # Throw away any fragment on the baseURL 
  regsub {#[^#]*$} $baseURL {} baseURL
  if [regexp {^[a-zA-Z0-9$_+.!*'(),;/?:@&=%~-]*$} $relativeURL] {
   if [regexp . $baseURL] {
    if [catch {OMFURL_ParseUrl $relativeURL} foo] {
     array set baseparts [OMFURL_ParseUrl $baseURL]
     if [catch {OMFURL_ParseIpSchemepart [list schemepart $relativeURL]} foo] {
      array set baseparts [OMFURL_ParseIpSchemepart [array get baseparts]]
      if {![regexp ^/ $relativeURL]} {
       regexp {^([^;?]*)?(;[^?]*)?(\?.*)?$} $baseparts(rel_path) match \
	baseparts(path) baseparts(params) baseparts(query)
       regexp {^([^;?]*)?(;[^?]*)?(\?.*)?$} $relativeURL match \
	URLparts(path) URLparts(params) URLparts(query)
       if [regexp . $URLparts(path)] {
	if [regexp / $baseparts(path)] {
	 regsub {/[^/]*$} $baseparts(path) / path
	 set path /$path$URLparts(path)
	} else {
	 set path /$URLparts(path)
	}
	while {[regsub {(/\./)|(/\.$)} $path / path]} {}
	while {[regsub [join { 
			(/[^/]/\.\./)|
			(/[^/][^/][^/]+/\.\./)|
			(/[^/][^/.]/\.\./)|
			(/[^/.][^/]/\.\./)|
			(/[^/]/\.\.$)|
			(/[^/][^/][^/]+/\.\.$)|
			(/[^/][^/.]/\.\.$)|
			(/[^/.][^/]/\.\.$)
		       } ""] $path / path]} {}
	regsub {^/} $path {} path
	set params $URLparts(params)
	set query $URLparts(query)
       } else {
	set path $baseparts(path)
	if [regexp . URLparts(params)] {
	 set params $URLparts(params)
	 set query $URLparts(query)
	} else {
	 set params $baseparts(params)
	 if [regexp . URLparts(query)] {
	  set query $URLparts(query)
	 } else {
	  set query $baseparts(query)
	 }
	}
       }
       return $baseparts(scheme)://$baseparts(net_loc)/$path$params$query
      } else {
       return $baseparts(scheme)://$baseparts(net_loc)$relativeURL
      }
     } else {
      return $baseparts(scheme):$relativeURL
     }
    } else {
     return $relativeURL
    }
   } else {
     return $relativeURL
   }
  } else {
    # There's an error -- relative URL contains unescaped illegal characters
    # Return unprocessed relative URL, and it will cause retrieval error.
    return -code error -erroinfo "Illegal characters in URL" $relativeURL
  }
}

# proc FileUrlToFile fileURL
#       Takes a URL of the file://... type and returns a file name which can
#       be read to get at the contents of the URL.
#
#       NOTE: Returns error messages in HTML
proc OMFURL_FileUrlToFile {fileURL} {
  global tcl_platform
  if [catch {OMFURL_ParseUrl $fileURL} URLpartlist] {
    return -code error "<P>Invalid URL syntax."
  }
  array set URLparts $URLpartlist
  if [catch {OMFURL_ParseIpSchemepart [array get URLparts]} URLpartlist] {
    return -code error "<P>Invalid URL syntax for protocol &quot;file&quot; ."
  }  
  array set URLparts $URLpartlist
  if [string compare $URLparts(net_loc) localhost] {
    return -code error "\
<p>Protocol &quot;file&quot; implemented only for host &quot;localhost&quot; .\
<p>Cannot use with host &quot;$URLparts(net_loc)&quot; .\
<p>Try a real WWW browser."
  }  
  set filename [OMFURL_UrlUnEscape $URLparts(abs_path)]
# On Windows we have to extract the volume information from the URL
  if [info exists tcl_platform] {
    if {![string compare $tcl_platform(platform) windows]} {
      regsub {^/([^/]+)} $filename {\1:} filename
    }
  }
  if [file isdirectory $filename] {
    set filename [file join $filename index.html]
  }
  return $filename
}

# proc FileToUrl filename
#       Takes a filename and returns the absolute URL of the named file.
proc OMFURL_FileToUrl {filename} {
  # If filename does not have absolute form, create its absolute form
  #     Let the cd and pwd commands do the work.  pwd returns an absolute path
  #     and cd accepts any legal directory name, so go where the file is
  #     and invoke pwd to get its parent directory.
  set saveworkingdirectory [pwd]
  cd [file dirname $filename]
  set filename [file join [pwd] [file tail $filename]]
  cd $saveworkingdirectory
  # Now filename has absolute form.
  # Split into its component pieces
  set fncomps [file split $filename]
  # Convert Windows and Mac root form to something URL suitable
  set root [lindex $fncomps 0]
  if [regexp {([^:]+):} $root match drive] {
    set fncomps [lreplace $fncomps 0 0 /$drive]
  }
  # URL escapes
  set apcomps {}
  foreach comp $fncomps {
    lappend apcomps [OMFURL_UrlEscape $comp]
  }
  set abspath [join $apcomps /]
  # Remove any double "/" at beginning of abspath
  while {[regsub {^//} $abspath {/} abspath]} {}
  # Slap a trailing / on the URL of a directory
  if [file isdirectory $filename] {
    return file://localhost$abspath/
  } else {
    return file://localhost$abspath
  }
}

# proc UrlEscape string
#       URLs are (strictly) only allowed to contain certain characters
#       Other characters are represented by escape sequences.  This
#       proc returns a URL-escaped form of its argument
proc OMFURL_UrlEscape {string} {
  # Escape everything that needs it, and dollar signs too (Tcl subst safety)
  regsub -all {[^a-zA-Z0-9_+.!*'(),;/?:@&=-]} $string \
	      {%[scan & %c v; format %x $v]} string
  set string [subst $string]
  # Then un-escape the dollar signs.
  regsub -all %[scan $ %c v; format %x $v] $string $ string
  return $string
}

# proc UrlUnEscape string
#       Reverses the effects of UrlEscape
proc OMFURL_UrlUnEscape {string} {
  # Convert all %hh to corresponding character
  regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $string \
	      {[scan \1 %x v; format %c $v]} string
  set string [subst $string]
  return $string
}

